"use strict";

exports.__esModule = true;
exports.default = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _akutils = require("akutils");

var _lodash = require("lodash.clonedeep");

var _lodash2 = _interopRequireDefault(_lodash);

var _lodash3 = require("lodash.defaults");

var _lodash4 = _interopRequireDefault(_lodash3);

var _lodash5 = require("lodash.isarray");

var _lodash6 = _interopRequireDefault(_lodash5);

var _lodash7 = require("lodash.isplainobject");

var _lodash8 = _interopRequireDefault(_lodash7);

var _Shadow = require("./Shadow");

var _Shadow2 = _interopRequireDefault(_Shadow);

var _ShadowImpl = require("./ShadowImpl");

var _ShadowImpl2 = _interopRequireDefault(_ShadowImpl);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/*
	Defines the shadowing behavior for a property. f.lux uses StateType instances to configure Shader
	instances that are then used to proxy the store's state using shadow properties.

	Property subclasses usually expose a 'type' class variable to specify their default shadowing behavior.
	The 'type' class variable should be a getter property and is typically attached to the class using the
	StateType.defineType() function.

	Example of defining a 'type' class variable:

		StateType.defineType(TodoRootProperty, spec => {
			spec.initialState({})                       // empty object initial state
				.autoshadowOff                          // do not shadow state values without explicit sub-property definitions
				.properties({                           // define sub-properties (just one in this case)
						todos: TodoCollection.type,     // 'todos' is a collection property
					})
				.readonly                               // prevent application code from reassigning the 'todos' collection (paranoia)
				.typeName("TodoRootProperty");          // useful for certain diagnostic situations
		});

	Shadowing behavior may be customized by setting the following StateType values:

		* autoshadow - shadow implicit state properties using default f.lux property types (default=true)
		* defaults - specifies default values for an object's initial state. A default value is applied
			when the initial state for this property is 'undefined'. (default=undefined)
		* element type - StateType used for shadowing each sub-property, such as elements in an array or
			properties of an object. (default=null)
		* implementation class - ShadowImpl subclass used to back shadow state property (default=null)
		* initial state - the initial property value. The store's state will take precedence and this value
			is used when the store's state is undefined for the specific property. (default=undefined)
		* managed type - StateType instance used by CollectionProperty to specify the property type for
			each model contained in the collection. Other collection types could utilize the same
			mechanism. (default=null)
		* properties - literal object with name/value pairs where name is the sub-property name and the
			value is a StateType instance describing the property shadowing behavior (default={})
		* shadow class - Shadow subclass used to represent a shadow state property. Built-in implementation
			classes set a default so only need to set when want to customize parent type default. (default=null)

	Methods/properties used when setting up a StateType description:

		* addProperty(name, type) - add a single property type where 'name' is the sub-property name and
			value is a StateType. Consider using StateType.properties() instead.
		* autoshadow - turn on auto-shadowing (default)
		* autoshadowOff - turn off auto-shadowing
		* default(state) - used with ObjectProperty and MapProperty sub-properties to set defaults for
			values. The default value is only applied if the initial state for this property is
			undefined.

			Example usage:
				stateType.properties({
					name: PrimitiveProperty.type.default(null),
				})

		* elementType(type) - StateType for shadowing each sub-property. Regularly used with IndexedProperty
			and ArrayProperty. Used with ObjectProperty and MapProperty when each value is the same type such
			as in a dictionary.
		* implementationClass(cls) - sets the ShadowImpl subclass to be used for backing the shadow state
			(should not be needed)
		* initialState(state) - value suitable for the property type: boolean, array, object,...
		* managedType(type) - StateType for shading managed properties (used with CollectionProperty)

			Example usage:
				stateType.managedType(TodoProperty.type)

		* properties(propTypes) - add multiple properties using a literal object with property-name/StateType
			key/value pairs

			Example usage:
				stateType.properties({
						completed: PrimitiveProperty.type.initialState(false),
						created: PrimitiveProperty.type.readonly,
						desc: PrimitiveProperty.type,
						id: PrimitiveProperty.type.readonly,
					})

		* readonly - set property to readonly
		* readonlyOff - set property to read/write (default unless parent property has set readonly to true
			since readonly property cascades through the state tree)
		* typeName(name) - set a label for the type. Useful for debugging situations where shadowing is
			not acting as expected.

			Example usage:

				const root = store.shadow;
				const rootProperty = root.$$();

				console.log(rootProperty.typeName())
				console.log(rootProperty.stateType().getTypeName())

	Instance of this class are rarely directly created and instead are created using the defineType()
	static function.
*/

var StateType = function () {
	function StateType() {
		var _arguments = arguments;

		_classCallCheck(this, StateType);

		(0, _akutils.assert)(function (a) {
			return a.not(_arguments.length, "StateType constructor takes no arguments");
		});

		this._PropertyClass = undefined;
		this._autoshadow = true;
		this._defaults = undefined;
		this._elementType = null;
		this._implementationClass = null;
		this._initialState = undefined;
		this._managedType = null;
		this._properties = {};
		this._shadowClass = null;
		this._typeName = null;

		// readonly is different than other instance variables as readonly state cascades down
		// to properties where not explicitly set to true or false
		this._readonly = undefined;
	}

	/*
 	Convenience class method to avoid using 'new StateType(...)' making chaining on the
 	newly created StateType easier to read.
 */


	_createClass(StateType, [{
		key: "addProperty",
		value: function addProperty(name, type) {
			var _this = this;

			(0, _akutils.assert)(function (a) {
				a.is(isKeyedPrototype(_this._PropertyClass), "PropertyClass must be a subclass of KeyedProperty");
				a.is(type instanceof StateType, "'type' parameter not StateType");
			});

			this._properties[name] = type;
		}
	}, {
		key: "default",
		value: function _default(state) {
			this._defaults = state;

			return this;
		}
	}, {
		key: "elementType",
		value: function elementType(type) {
			(0, _akutils.assert)(function (a) {
				return a.is(type instanceof StateType, "'type' parameter not StateType");
			});

			this._elementType = type;

			return this;
		}
	}, {
		key: "implementationClass",
		value: function implementationClass(cls) {
			this._implementationClass = cls;

			return this;
		}
	}, {
		key: "initialState",
		value: function initialState(state) {
			this._initialState = state;

			return this;
		}
	}, {
		key: "managedType",
		value: function managedType(type) {
			(0, _akutils.assert)(function (a) {
				return a.is(type instanceof StateType, "'type' parameter not StateType");
			});

			this._managedType = type;

			return this;
		}
	}, {
		key: "propertyClass",
		value: function propertyClass(PropertyClass) {
			this._PropertyClass = PropertyClass;

			return this;
		}
	}, {
		key: "properties",
		value: function properties(propTypes) {
			var _this2 = this;

			(0, _akutils.assert)(function (a) {
				a.is(isKeyedPrototype(_this2._PropertyClass), "PropertyClass must subclass ObjectProperty or implement supportsKeyedChildProperties()");
			});

			for (var name in propTypes) {
				this.addProperty(name, propTypes[name]);
			}

			return this;
		}
	}, {
		key: "shadowClass",
		value: function shadowClass(cls) {
			this._shadowClass = cls;

			return this;
		}
	}, {
		key: "typeName",
		value: function typeName(name) {
			this._typeName = name;
			this._PropertyClass.__fluxTypeName__ = name;

			return this;
		}

		//---------------------------------------------------------------------------------------------
		//  StateType runtime API used during the shadowing process
		//---------------------------------------------------------------------------------------------

	}, {
		key: "clone",
		value: function clone() {
			return Object.assign(new StateType(), this);
		}
	}, {
		key: "computeInitialState",
		value: function computeInitialState() {
			var state = (0, _lodash2.default)(this._initialState);
			var propSpecs = this._properties;

			if (!(0, _lodash8.default)(state) || !propSpecs) {
				return state;
			}

			var propType, propState;

			for (var name in propSpecs) {
				propType = propSpecs[name];
				propState = state[name];

				// update state only if current undefined
				state[name] = propState === undefined ? propType.computeInitialState() : propState;
			}

			return state;
		}

		/*
  	Configures an existing shader with the values of this type.
  */

	}, {
		key: "configureShader",
		value: function configureShader(shader) {
			shader.setAutoshadow(this._autoshadow);

			this._setupShader(shader);
		}
	}, {
		key: "createProperty",
		value: function createProperty() {
			return new this._PropertyClass(this);
		}

		/*
  	Returns a PropertyFactoryShader that will create property instances as configured by this StateType.
  */

	}, {
		key: "factory",
		value: function factory(parentProperty) {
			var PropertyFactoryShader = require("./PropertyFactoryShader").default;
			var shader = new PropertyFactoryShader(this, parentProperty);

			this._setupShader(shader);

			return shader;
		}
	}, {
		key: "getManagedType",
		value: function getManagedType() {
			return this._managedType;
		}
	}, {
		key: "getTypeName",
		value: function getTypeName() {
			return this.ProperClass.__fluxTypeName__;
		}
	}, {
		key: "initialStateWithDefaults",
		value: function initialStateWithDefaults(state) {
			var arrayType = (0, _lodash6.default)(state);
			var propSpecs = this._properties;
			var defaultState, propType, propState;

			if (!(0, _lodash8.default)(state) && !arrayType) {
				return state;
			}

			if ((0, _lodash8.default)(this._defaults)) {
				state = (0, _lodash4.default)(arrayType ? [] : {}, state, this._defaults);
			}

			for (var name in propSpecs) {
				propType = propSpecs[name];

				if (propType && propType._defaults !== undefined) {
					defaultState = defaultState || (arrayType ? [] : {});
					defaultState[name] = propType._defaults;
				}
			}

			if (defaultState) {
				state = (0, _lodash4.default)(arrayType ? [] : {}, state, defaultState);
			}

			return state;
		}
	}, {
		key: "isComplex",
		value: function isComplex() {
			return Object.keys(this._properties).length || this._managedType || this._elementType;
		}
	}, {
		key: "shader",
		value: function shader(property) {
			var Shader = require("./Shader").default;
			var shader = new Shader(property);

			this._setupShader(shader);

			return shader;
		}
	}, {
		key: "shadowClassForProperty",
		value: function shadowClassForProperty() {
			var defaultClass = arguments.length <= 0 || arguments[0] === undefined ? _Shadow2.default : arguments[0];

			if (this._shadowClass) {
				return this._shadowClass;
			} else {
				return defaultClass;
			}
		}
	}, {
		key: "_setupShader",
		value: function _setupShader(shader) {
			if (this._elementType) {
				var eltType = this._elementType;

				shader.setElementType(eltType);
			} else {
				var _eltType = void 0;

				for (var name in this._properties) {
					_eltType = this._properties[name];

					shader.addProperty(name, _eltType);
				}
			}
		}
	}, {
		key: "autoshadow",


		//---------------------------------------------------------------------------------------------
		//  StateType setup API
		//---------------------------------------------------------------------------------------------

		get: function get() {
			this._autoshadow = true;

			return this;
		}
	}, {
		key: "autoshadowOff",
		get: function get() {
			this._autoshadow = false;

			return this;
		}
	}, {
		key: "readonly",
		get: function get() {
			this._readonly = true;

			return this;
		}
	}, {
		key: "readonlyOff",
		get: function get() {
			this._readonly = false;

			return this;
		}
	}], [{
		key: "create",
		value: function create(PropertyClass) {
			return new StateType().propertyClass(PropertyClass);
		}

		/*
  	Sets a 'type' getter on the PropertyClass. The 'type' getter is used to define properties
  	in the StateType.properties() method. The 'type' getter may be set once per class and all
  	further requests are ignored.
  		Example usage:
  			StateType.defineType(CounterProperty);
  			...
  			ObjectProperty.createClass({}, spec => {
  			spec.properties({
  					counter: CounterProperty.type.readonly,
  				})
  		});
  		Parameters:
  		PropertyClass - the Property subclass on which to define the 'type' getter
  		typeCallback - callback of form cb(spec) that allows the StateType to be customized
  */

	}, {
		key: "defineType",
		value: function defineType(PropertyClass, typeCallback) {
			(0, _akutils.assert)(function (a) {
				return a.not(PropertyClass.hasOwnProperty("type"), "'type' variable already defined");
			});

			if (PropertyClass.hasOwnProperty("type")) {
				return;
			}

			/*
   	Create the basis for the returned 'type'. The type getter needs to return a copy to permit
   	modifying the shadowing behavior.
   */
			var type = new StateType();

			type.propertyClass(PropertyClass);

			if (typeCallback) {
				typeCallback(type);
			}

			Object.defineProperty(PropertyClass, "type", {
				get: function get() {
					// permit modifications with altering the default case
					return type.clone();
				}
			});
		}

		/*
  	Walks the prototype chain and returns the first StateType returned by a 'type' class variable.
  		Parameters:
  		prop - one of a Property instance or a Property class prototype
  		Returns the StateType describing the property
  */

	}, {
		key: "from",
		value: function from(prop) {
			var ctor = prop.constructor;

			/*
   	Get the prototype of prop. two cases
   		1) prop is Property instance - proto will be its prototype
   		2) prop is actually a prototype - proto will then be the parent prototype
   */
			var proto = Object.getPrototypeOf(prop);

			return ctor.type || StateType.from(ctor === proto.constructor ? Object.getPrototypeOf(proto) : proto);
		}

		/*
  	Determines the ShadowImpl subclass to use for the property.
  */

	}, {
		key: "implementationClassForProperty",
		value: function implementationClassForProperty(property) {
			var defaultClass = arguments.length <= 1 || arguments[1] === undefined ? _ShadowImpl2.default : arguments[1];

			var stateType = StateType.from(property);

			if (stateType && stateType._implementationClass) {
				return stateType._implementationClass;
			} else {
				return defaultClass;
			}
		}
	}]);

	return StateType;
}();

exports.default = StateType;


function isKeyedPrototype(obj) {
	var ObjectProperty = require("./ObjectProperty").default;
	var Property = require("./Property").default;

	return ObjectProperty === obj || ObjectProperty.isPrototypeOf(obj) || Property.isPrototypeOf(obj) && obj.supportsKeyedChildProperties && obj.supportsKeyedChildProperties();
}